@@ -25,6 +25,11 @@ if (&read_file("$mcd/site", \%site)) {
 	foreach $m (split(/\s+/, $site{'modules'})) {
 		if ($m =~ /(\S+)\/(\S+)/) { $httpd_modules{$1} = $2; }
 		}
+	foreach $m (keys %httpd_modules) {
+		if (!-r "$module_root_directory/$m.pl") {
+			delete($httpd_modules{$m});
+			}
+		}
 	foreach $f (split(/\s+/, $site{'htaccess'})) {
 		if (-r $f) { push(@htaccess_files, $f); }
 		}
@@ -816,8 +821,7 @@ sub restart_button
 local $rv;
 local $pidfile = &get_pid_file();
 $args = "redir=".&urlize(&this_url());
-if (open(PID, $pidfile) && <PID> =~ /(\d+)/ && kill(0, $1)) {
-	close(PID);
+if (&check_pid_file($pidfile)) {
 	if ($access{'apply'}) {
 		$rv .= "<a href=\"restart.cgi?$args\">$text{'apache_apply'}</a><br>\n";
 		}
@@ -1131,5 +1135,45 @@ local $pidfile = $pidfilestr ? $pidfilestr->{'words'}->[0]
 return &server_root($pidfile, $conf);
 }
 
+# restart_apache()
+# Re-starts Apache, and returns undef on success or an error message on failure
+sub restart_apache
+{
+local $conf = &get_config();
+local $pidfile = &get_pid_file();
+if ($config{'apply_cmd'}) {
+	# use the configured start command
+	local $out = &backquote_logged("($config{'apply_cmd'}) 2>&1");
+	if ($?) {
+		return "<pre>$out</pre>";
+		}
+	}
+elsif (-x $config{'apachectl_path'}) {
+	# use apachectl to restart
+	if ($httpd_modules{'core'} >= 2) {
+		# Do a graceful restart
+		local $out = &backquote_logged("$config{'apachectl_path'} graceful 2>&1");
+		if ($?) {
+			return "<pre>$out</pre>";
+			}
+		}
+	else {
+		local $out = &backquote_logged("$config{'apachectl_path'} restart 2>&1");
+		if ($out !~ /httpd restarted/) {
+			return "<pre>$out</pre>";
+			}
+		}
+	}
+else {
+	# send SIGHUP directly
+	open(PID, $pidfile) || return &text('restart_epid', $pidfile);
+	<PID> =~ /(\d+)/ || return &text('restart_epid2', $pidfile);
+	close(PID);
+	&kill_logged('HUP', $1) || return &text('restart_esig', $1);
+	}
+return undef;
+
+}
+
 1;
 
