@@ -2,6 +2,10 @@
 # XXX don't connect when viewing cached mail?
 # XXX can we cache IMAP messages?
 # XXX don't reply-to our address
+# XXX dsns and multiple recipients
+# 	XXX and delivery notifications
+#	XXX put in separate hashes?
+#	XXX existing code to detect bounces is wrong
 
 do '../web-lib.pl';
 &init_config();
@@ -155,16 +159,16 @@ local $mode = @_ > 2 ? $_[2] : 0;
 local $nogroups = @_ > 4 ? $_[4] : 0;
 local ($rfield1, $rfield2);
 if ($_[3]) {
-	return "<input type=button onClick='ifield = document.forms[$form].$_[0]; rfield = document.forms[$form].$_[3]; chooser = window.open(\"address_chooser.cgi?addr=\"+escape(ifield.value)+\"&mode=$mode&nogroups=$nogroups\", \"chooser\", \"toolbar=no,menubar=no,scrollbars=yes,width=500,height=250\"); chooser.ifield = ifield; window.ifield = ifield; chooser.rfield = rfield; window.rfield = rfield' value=\"...\">\n";
+	return "<input type=button onClick='ifield = document.forms[$form].$_[0]; rfield = document.forms[$form].$_[3]; chooser = window.open(\"../$module_name/address_chooser.cgi?addr=\"+escape(ifield.value)+\"&mode=$mode&nogroups=$nogroups\", \"chooser\", \"toolbar=no,menubar=no,scrollbars=yes,width=500,height=250\"); chooser.ifield = ifield; window.ifield = ifield; chooser.rfield = rfield; window.rfield = rfield' value=\"...\">\n";
 	}
 else {
-	return "<input type=button onClick='ifield = document.forms[$form].$_[0]; chooser = window.open(\"address_chooser.cgi?addr=\"+escape(ifield.value)+\"&mode=$mode\", \"chooser\", \"toolbar=no,menubar=no,scrollbars=yes,width=500,height=250\"); chooser.ifield = ifield; window.ifield = ifield' value=\"...\">\n";
+	return "<input type=button onClick='ifield = document.forms[$form].$_[0]; chooser = window.open(\"../$module_name/address_chooser.cgi?addr=\"+escape(ifield.value)+\"&mode=$mode\", \"chooser\", \"toolbar=no,menubar=no,scrollbars=yes,width=500,height=250\"); chooser.ifield = ifield; window.ifield = ifield' value=\"...\">\n";
 	}
 }
 
 # list_folders()
 # Returns a list of all folders for this user
-# folder types: 0 = mbox, 1 = maildir, 2 = pop3, 3 = mh, 4 = imap
+# folder types: 0 = mbox, 1 = maildir, 2 = pop3, 3 = mh, 4 = imap, 5 = combined
 # folder modes: 0 = ~/mail, 1 = external folder, 2 = sent mail,
 #               3 = inbox/drafts/trash
 sub list_folders
@@ -301,6 +305,26 @@ foreach $f (readdir(DIR)) {
 		}
 	}
 closedir(DIR);
+
+# Add user-defined composite folders
+opendir(DIR, $user_module_config_directory);
+foreach $f (readdir(DIR)) {
+	if ($f =~ /^(\d+)\.comp$/) {
+		local %comp;
+		&read_file("$user_module_config_directory/$f", \%comp);
+		$comp{'type'} = 5;
+		$comp{'mode'} = 0;
+		$comp{'index'} = scalar(@rv);
+		local $sfn;
+		foreach $sfn (split(/\t+/, $comp{'subfoldernames'})) {
+			local $sf = &find_named_folder($sfn, \@rv);
+			push(@{$comp{'subfolders'}}, $sf) if ($sf);
+			}
+		push(@rv, \%comp);
+		}
+	}
+closedir(DIR);
+
 foreach $f (@rv) {
 	if ($f->{'file'} && $userconfig{"notes_".$f->{'file'}}) {
 		$f->{'notes_decode'} = 1;
@@ -389,9 +413,9 @@ sub movecopy_select
 {
 local $rv;
 $rv .= "<input type=submit name=move$_[0] value=\"$text{'mail_move'}\" ".
-       "onClick='return check_clicks(form)'>\n";
+       "onClick='return check_clicks(form)'>";
 $rv .= "<input type=submit name=copy$_[0] value=\"$text{'mail_copy'}\" ".
-       "onClick='return check_clicks(form)'>\n";
+       "onClick='return check_clicks(form)'>";
 local @mfolders = grep { $_ ne $_[2] && !$_->{'nowrite'} } @{$_[1]};
 $rv .= &folder_select(\@mfolders, undef, "mfolder$_[0]");
 return $rv;
@@ -513,13 +537,14 @@ elsif ($userconfig{'folder_sort'} == 2) {
 }
 
 # find_named_folder(name, &folders)
-# Finds a folder by filename, server name or displayed name
+# Finds a folder by ID, filename, server name or displayed name
 sub find_named_folder
 {
-local ($rv) = grep { $_->{'file'} eq $_[0] ||
-		     $_->{'server'} eq $_[0] } @{$_[1]};
-($rv) = grep { $_->{'name'} eq $_[0] } @{$_[1]}
-	if (!$rv);
+local $rv;
+($rv) = grep { $_->{'id'} eq $_[0] } @{$_[1]} if (!$rv);
+($rv) = grep { $_->{'file'} eq $_[0] ||
+		     $_->{'server'} eq $_[0] } @{$_[1]} if (!$rv);
+($rv) = grep { $_->{'name'} eq $_[0] } @{$_[1]} if (!$rv);
 return $rv;
 }
 
@@ -563,12 +588,14 @@ $module_index_name = $text{'mail_indexlink'};
 sub check_modification
 {
 local $newmod = &modification_time($_[0]);
-if ($in{'mod'} && $in{'mod'} != $newmod) {
+if ($in{'mod'} && $in{'mod'} != $newmod && $userconfig{'check_mod'}) {
 	# Changed!
 	&error(&text('emodified', "index.cgi?folder=$_[0]->{'index'}"));
 	}
 }
 
+# list_from_addresses()
+# Returns a list of allowed From: addresses for the current user
 sub list_from_addresses
 {
 local $http_host = $ENV{'HTTP_HOST'};
@@ -596,9 +623,17 @@ if ($config{'from_map'}) {
 		s/\r|\n//g;
 		s/#.*$//;
 		if (/^\s*(\S+)\s+(\S+)/ &&
-		    ($1 eq $remote_user || &indexof($1, @froms) >= 0)) {
+		    ($1 eq $remote_user || &indexof($1, @froms) >= 0) &&
+		    $config{'from_format'} == 0) {
+			# Username on LHS matches
 			push(@mfroms, $2);
 			}
+		elsif (/^\s*(\S+)\s+(\S+)/ &&
+		    ($2 eq $remote_user || &indexof($2, @froms) >= 0) &&
+		    $config{'from_format'} == 1) {
+			# Username on RHS matches
+			push(@mfroms, $1);
+			}
 		}
 	close(MAP);
 	}
@@ -612,5 +647,108 @@ foreach $f (@froms) {
 return (\@froms, \@doms);
 }
 
+# update_delivery_notification(&mail, &folder)
+# If the given mail is a DSN, update the original email so we know it has
+# been read
+sub update_delivery_notification
+{
+local ($mail, $folder) = @_;
+return 0 if ($mail->{'header'}->{'content-type'} !~ /multipart\/report/i);
+local $mid = $mail->{'header'}->{'message-id'};
+&open_dsn_hash();
+if ($dsnreplies{$mid} || $delreplies{$mid}) {
+	# We have already done this DSN
+	return 0;
+	}
+if (!defined($mail->{'body'}) && !$mail->{'parsed'} &&
+    defined($mail->{'idx'})) {
+	# This message has no body, perhaps because one wasn't fetched ..
+	local @mail = &mailbox_list_mails($mail->{'idx'}, $mail->{'idx'},
+					  $folder);
+	$mail = $mail[$mail->{'idx'}];
+	}
+$dsnreplies{$mid} = $delreplies{$mid} = 1;
+
+# Find the delivery or disposition status attachment
+&parse_mail($mail);
+local ($dsnattach) = grep { $_->{'header'}->{'content-type'} =~ /message\/disposition-notification/i } @{$mail->{'attach'}};
+local ($delattach) = grep { $_->{'header'}->{'content-type'} =~ /message\/delivery-status/i } @{$mail->{'attach'}};
+
+if ($dsnattach) {
+	# Update the read status for the original message
+	if ($dsnattach->{'data'} =~ /Original-Message-ID:\s*(.*)/) {
+		$omid = $1;
+		}
+	else {
+		return 0;
+		}
+	local ($faddr) = &split_addresses($mail->{'header'}->{'from'});
+	&add_address_to_hash(\%dsnreplies, $omid, $faddr->[0]);
+	return 1;
+	}
+elsif ($delattach) {
+	# Update the delivery status for the original message, which will be
+	# in a separate attachment
+	local ($origattach) = grep { $_->{'header'}->{'content-type'} =~ /text\/rfc822-headers|message\/rfc822/i } @{$mail->{'attach'}};
+	return 0 if (!$origattach);
+	local $origmail = &extract_mail($origattach->{'data'});
+	local $omid = $origmail->{'header'}->{'message-id'};
+	return 0 if (!$omid);
+	local ($faddr) = &split_addresses($origmail->{'header'}->{'from'});
+	local $ds = &parse_delivery_status($delattach->{'data'});
+	if ($ds->{'status'} =~ /^2\./) {
+		&add_address_to_hash(\%delreplies, $omid, $faddr->[0]);
+		}
+	elsif ($ds->{'status'} =~ /^5\./) {
+		&add_address_to_hash(\%delreplies, $omid, "!".$faddr->[0]);
+		}
+	}
+else {
+	return 0;
+	}
+}
+
+# add_address_to_hash(&hash, messageid, address)
+sub add_address_to_hash
+{
+local @cv = split(/\s+/, $_[0]->{$_[1]});
+local $idx = &indexof($_[2], @cv);
+if ($idx < 0) {
+	$_[0]->{$_[1]} .= " " if (@cv);
+	$_[0]->{$_[1]} .= time()." ".$_[2];
+	}
+}
+
+# open_dsn_hash()
+# Ensure the %dsnreplies and %delreplies hashes are tied
+sub open_dsn_hash
+{
+if (!defined(%dsnreplies)) {
+	dbmopen(%dsnreplies, "$user_module_config_directory/dsnreplies", 0600);
+	}
+if (!defined(%delreplies)) {
+	dbmopen(%delreplies, "$user_module_config_directory/delreplies", 0600);
+	}
+}
+
+# spam_report_cmd()
+# Returns a command for reporting spam, or undef if none
+sub spam_report_cmd
+{
+local %sconfig = &foreign_config("spam");
+if ($config{'spam_report'} eq 'sa_learn') {
+	return &has_command($sconfig{'sa_learn'}) ? "$sconfig{'sa_learn'} --spam" : undef;
+	}
+elsif ($config{'spam_report'} eq 'spamassassin') {
+	return &has_command($sconfig{'spamassassin'}) ? "$sconfig{'spamassassin'} --r" : undef;
+	}
+else {
+	return &has_command($sconfig{'sa_learn'}) ?
+		"$sconfig{'sa_learn'} --spam" :
+	       &has_command($sconfig{'spamassassin'}) ?
+		"$sconfig{'spamassassin'} --r" : undef;
+	}
+}
+
 1;
 
