@@ -26,5 +26,12 @@ if ($perl_doc && $] >= 5.006) {
 	$perl_doc = "$perl_doc -U";
 	}
 
+# set_manpath(extra)
+sub set_manpath
+{
+$ENV{'MANPATH'} = join(":", split(/:/, $config{'man_dir'}),
+			    split(/:/, $_[0]));
+}
+
 1;
 
