@@ -86,7 +86,7 @@ close(SOURCE);
 	  "pam", "webmin",
 	  "sidname", "sid",
 	  "unauth", "^/unauthenticated/ ^[A-Za-z0-9\\-/]+\\.jar\$ ^[A-Za-z0-9\\-/]+\\.class\$ ^[A-Za-z0-9\\-/]+\\.gif\$ ^[A-Za-z0-9\\-/]+\\.conf\$",
-	  "max_post", 10000
+	  "max_post", 10000,
 	 );
 foreach $v (keys %vital) {
 	if (!$config{$v}) {
@@ -149,7 +149,11 @@ if ($config{'pam_only'} && !$use_pam) {
 	   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" );
 
 # Change dir to the server root
-chdir($config{'root'});
+@roots = ( $config{'root'} );
+for($i=0; defined($config{"extraroot_$i"}); $i++) {
+	push(@roots, $config{"extraroot_$i"});
+	}
+chdir($roots[0]);
 $user_homedir = (getpwuid($<))[7];
 
 # Read users file
@@ -207,7 +211,10 @@ if ($use_ssl) {
 
 # Setup syslog support if possible and if requested
 if ($use_syslog) {
-	eval 'openlog($config{"pam"}, "cons,pid,ndelay", "authpriv"); setlogsock("unix")';
+	open(ERRDUP, ">&STDERR");
+	open(STDERR, ">/dev/null");
+	$log_socket = $config{"logsock"} || "unix";
+	eval 'openlog($config{"pam"}, "cons,pid,ndelay", "authpriv"); setlogsock($log_socket)';
 	if ($@) {
 		$use_syslog = 0;
 		}
@@ -225,6 +232,7 @@ if ($use_syslog) {
 				}
 			}
 		}
+	open(STDERR, ">&ERRDUP");
 	}
 
 # Read MIME types file and add extra types
@@ -249,8 +257,6 @@ foreach $k (keys %config) {
 # get the time zone
 if ($config{'log'}) {
 	local(@gmt, @lct, $days, $hours, $mins);
-	@make_date_marr = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
-		 	   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
 	@gmt = gmtime(time());
 	@lct = localtime(time());
 	$days = $lct[3] - $gmt[3];
@@ -294,9 +300,7 @@ if ($config{'extauth'}) {
 		close(STDOUT);
 		open(STDIN, "<&EXTAUTH2");
 		open(STDOUT, ">&EXTAUTH2");
-		exec($config{'extauth'});
-		print STDERR "exec failed : $!\n";
-		exit 1;
+		exec($config{'extauth'}) or die "exec failed : $!\n";
 		}
 	close(EXTAUTH2);
 	local $os = select(EXTAUTH);
@@ -484,6 +488,8 @@ if ($config{'session'}) {
 $SIG{'HUP'} = 'miniserv::trigger_restart';
 $SIG{'TERM'} = 'miniserv::term_handler';
 $SIG{'PIPE'} = 'IGNORE';
+local $remove_session_count = 0;
+$need_pipes = $config{'passdelay'} || $config{'session'};
 while(1) {
 	# wait for a new connection, or a message from a child process
 	local ($i, $rmask);
@@ -498,7 +504,7 @@ while(1) {
 		printf STDERR "too many children (%d > %d)\n",
 			scalar(@childpids), $config{'maxconns'};
 		}
-	if ($config{'passdelay'} || $config{'session'}) {
+	if ($need_pipes) {
 		for($i=0; $i<@passin; $i++) {
 			vec($rmask, fileno($passin[$i]), 1) = 1;
 			}
@@ -558,7 +564,7 @@ while(1) {
 
 			# create pipes
 			local ($PASSINr, $PASSINw, $PASSOUTr, $PASSOUTw);
-			if ($config{'passdelay'} || $config{'session'}) {
+			if ($need_pipes) {
 				local $p;
 				local %taken = map { $_, 1 } @passin;
 				for($p=0; $taken{"PASSINr$p"}; $p++) { }
@@ -632,7 +638,7 @@ while(1) {
 					}
 
 				# close useless pipes
-				if ($config{'passdelay'} || $config{'session'}) {
+				if ($need_pipes) {
 					local $p;
 					foreach $p (@passin) { close($p); }
 					foreach $p (@passout) { close($p); }
@@ -648,7 +654,7 @@ while(1) {
 				exit;
 				}
 			push(@childpids, $handpid);
-			if ($config{'passdelay'} || $config{'session'}) {
+			if ($need_pipes) {
 				close($PASSINw); close($PASSOUTr);
 				push(@passin, $PASSINr); push(@passout, $PASSOUTw);
 				}
@@ -1425,10 +1431,10 @@ local $preroot = $authuser && defined($config{'preroot_'.$authuser}) ?
 if ($preroot) {
 	# Always under the current webmin root
 	$preroot =~ s/^.*\///g;
-	$preroot = $config{'root'}.'/'.$preroot;
+	$preroot = $roots[0].'/'.$preroot;
 	}
 if ($preroot) {
-	# Look in the template root directory first
+	# Look in the theme root directory first
 	$is_directory = 1;
 	$sofar = "";
 	$full = $preroot.$sofar;
@@ -1457,13 +1463,13 @@ if ($preroot) {
 		}
 	if ($full) {
 		if ($sofar eq '') {
-			$cgi_pwd = $config{'root'};
+			$cgi_pwd = $roots[0];
 			}
 		elsif ($is_directory) {
-			$cgi_pwd = "$config{'root'}$sofar";
+			$cgi_pwd = "$roots[0]$sofar";
 			}
 		else {
-			"$config{'root'}$sofar" =~ /^(.*\/)[^\/]+$/;
+			"$roots[0]$sofar" =~ /^(.*\/)[^\/]+$/;
 			$cgi_pwd = $1;
 			}
 		if ($is_directory) {
@@ -1486,31 +1492,47 @@ if ($preroot) {
 			}
 		}
 	}
-if (!$full || $is_directory) {
-	$sofar = "";
-	$full = $config{"root"} . $sofar;
-	$scriptname = $simple;
-	foreach $b (split(/\//, $simple)) {
-		if ($b ne "") { $sofar .= "/$b"; }
-		$full = $config{"root"} . $sofar;
-		@stfull = stat($full);
-		if (!@stfull) { &http_error(404, "File not found"); }
 
-		# Check if this is a directory
-		if (-d _) {
-			# It is.. go on parsing
-			next;
+# Look in the real root directories, stopping when we find a file or directory
+if (!$full || $is_directory) {
+	ROOT: foreach $root (@roots) {
+		$sofar = "";
+		$full = $root.$sofar;
+		$scriptname = $simple;
+		foreach $b ($simple eq "/" ? ( "" ) : split(/\//, $simple)) {
+			if ($b ne "") { $sofar .= "/$b"; }
+			$full = $root.$sofar;
+			@stfull = stat($full);
+			if (!@stfull) {
+				next ROOT;
+				}
+
+			# Check if this is a directory
+			if (-d _) {
+				# It is.. go on parsing
+				next;
+				}
+
+			# Check if this is a CGI program
+			if (&get_type($full) eq "internal/cgi") {
+				$pathinfo = substr($simple, length($sofar));
+				$pathinfo .= "/" if ($page =~ /\/$/);
+				$scriptname = $sofar;
+				last;
+				}
 			}
 
-		# Check if this is a CGI program
-		if (&get_type($full) eq "internal/cgi") {
-			$pathinfo = substr($simple, length($sofar));
-			$pathinfo .= "/" if ($page =~ /\/$/);
-			$scriptname = $sofar;
+		# Run CGI in the same directory as whatever file
+		# was requested
+		$full =~ /^(.*\/)[^\/]+$/; $cgi_pwd = $1;
+
+		if (-e $full) {
+			# Found something!
+			$realroot = $root;
 			last;
 			}
 		}
-	$full =~ /^(.*\/)[^\/]+$/; $cgi_pwd = $1;
+	if (!@stfull) { &http_error(404, "File not found"); }
 	}
 @stfull = stat($full) if (!@stfull);
 
@@ -1605,7 +1627,8 @@ if (&get_type($full) eq "internal/cgi") {
 	$ENV{"SERVER_SOFTWARE"} = $config{"server"};
 	$ENV{"SERVER_NAME"} = $host;
 	$ENV{"SERVER_ADMIN"} = $config{"email"};
-	$ENV{"SERVER_ROOT"} = $config{"root"};
+	$ENV{"SERVER_ROOT"} = $roots[0];
+	$ENV{"SERVER_REALROOT"} = $realroot;
 	$ENV{"SERVER_PORT"} = $port;
 	$ENV{"REMOTE_HOST"} = $acpthost;
 	$ENV{"REMOTE_ADDR"} = $acptip;
@@ -1615,7 +1638,8 @@ if (&get_type($full) eq "internal/cgi") {
 					    $config{'pass_password'});
 	$ENV{"SSL_USER"} = $peername if ($validated == 2);
 	$ENV{"ANONYMOUS_USER"} = "1" if ($validated == 3);
-	$ENV{"DOCUMENT_ROOT"} = $config{"root"};
+	$ENV{"DOCUMENT_ROOT"} = $roots[0];
+	$ENV{"DOCUMENT_REALROOT"} = $realroot;
 	$ENV{"GATEWAY_INTERFACE"} = "CGI/1.1";
 	$ENV{"SERVER_PROTOCOL"} = "HTTP/1.0";
 	$ENV{"REQUEST_METHOD"} = $method;
@@ -1623,7 +1647,8 @@ if (&get_type($full) eq "internal/cgi") {
 	$ENV{"SCRIPT_FILENAME"} = $full;
 	$ENV{"REQUEST_URI"} = $request_uri;
 	$ENV{"PATH_INFO"} = $pathinfo;
-	$ENV{"PATH_TRANSLATED"} = "$config{root}/$pathinfo";
+	$ENV{"PATH_TRANSLATED"} = "$roots[0]/$pathinfo";
+	$ENV{"PATH_REALTRANSLATED"} = "$realroot/$pathinfo";
 	$ENV{"QUERY_STRING"} = $querystring;
 	$ENV{"MINISERV_CONFIG"} = $conf;
 	$ENV{"HTTPS"} = "ON" if ($use_ssl || $config{'inetd_ssl'});
@@ -1728,9 +1753,8 @@ if (&get_type($full) eq "internal/cgi") {
 			open(STDOUT, ">&CGIOUTw");
 			open(STDERR, ">&CGIERRw");
 			close(CGIINw); close(CGIOUTr); close(CGIERRr);
-			exec($full, split(/\s+/, $queryargs));
-			print STDERR "Failed to exec $full : $!\n";
-			exit;
+			exec($full, split(/\s+/, $queryargs))
+				or die "Failed to exec $full : $!\n";
 			}
 		close(CGIINr); close(CGIOUTw); close(CGIERRw);
 
@@ -2168,7 +2192,7 @@ sub make_datestr
 {
 local @tm = localtime(time());
 return sprintf "%2.2d/%s/%4.4d:%2.2d:%2.2d:%2.2d %s",
-		$tm[3], $make_date_marr[$tm[4]], $tm[5]+1900,
+		$tm[3], $month[$tm[4]], $tm[5]+1900,
 	        $tm[2], $tm[1], $tm[0], $timezone;
 }
 
@@ -2247,7 +2271,9 @@ sub PRINT
 {
 $r = shift;
 $$r++;
-&write_to_sock(@_);
+my $buf = join(defined($,) ? $, : "", @_);
+$buf .= $\ if defined($\);
+&write_to_sock($buf);
 }
  
 sub PRINTF
@@ -2416,7 +2442,12 @@ if (!$users{$_[0]}) {
 			s/\r|\n//g;
 			s/#.*$//;
 			if (/^(\S+)\s+(\S+)/) {
-				$user_mapping{$2} = $1;
+				if ($config{'user_mapping_reverse'}) {
+					$user_mapping{$1} = $2;
+					}
+				else {
+					$user_mapping{$2} = $1;
+					}
 				}
 			}
 		close(MAPPING);
@@ -2540,6 +2571,10 @@ if ($use_pam) {
 			       $acct_ret == PAM_ACCT_EXPIRED()) {
 				return 2;
 				}
+			else {
+				print STDERR "Unknown pam_acct_mgmt return value : $acct_ret\n";
+				return 2;
+				}
 			}
 		return 0;
 		}
@@ -2649,7 +2684,7 @@ sub get_socket_name
 return $config{'host'} if ($config{'host'});
 local $sn = getsockname($_[0]);
 return undef if (!$sn);
-local ($myport, $myaddr) = unpack_sockaddr_in($sn);
+local $myaddr = (unpack_sockaddr_in($sn))[1];
 if (!$get_socket_name_cache{$myaddr}) {
 	local $myname = gethostbyaddr($myaddr, AF_INET);
 	if ($myname eq "") {
