@@ -273,7 +273,7 @@ if ($remote_user && @_ > 1) {
 	}
 
 local $one = @msc_modules == 1 && $gconfig{'gotoone'};
-if (@_ > 1 && !$one && $remote_user) {
+if (@_ > 1 && !$one && $remote_user && !$gconfig{'notabs'}) {
     # Display module categories
     print qq~<table width="100%" border="0" cellspacing="0" cellpadding="0" height="7">
   <tr>
@@ -324,7 +324,7 @@ if (@_ > 1 && !$one && $remote_user) {
 		}
     }
 
-#####Navigation Bar START#####
+    # Display module categories
     print qq~<table width="100%" border="0" cellspacing="0" cellpadding="0" height="57" background="$gconfig{'webprefix'}/images/nav/bg.jpg">
   <tr background="$gconfig{'webprefix'}/images/nav/bg.jpg">
     <td width="6" nowrap><img src="$gconfig{'webprefix'}/images/nav/left.jpg" width="3" height="57"></td>~;
@@ -442,7 +442,7 @@ if (@_ > 1 && (!$_[5] || $ENV{'HTTP_WEBMIN_SERVERS'})) {
 		if (!$access{'noconfig'} && !$config{'noprefs'}) {
 			local $cprog = $user_module_config_directory ?
 					"uconfig.cgi" : "config.cgi";
-			&tab_start; print "<a href=\"/$cprog?$module_name\">",
+			&tab_start; print "<a href=\"$gconfig{'webprefix'}/$cprog?$module_name\">",
 			      $text{'header_config'},"</a>"; &tab_end;
 			}
 		}
@@ -661,9 +661,10 @@ return $rv;
 
 sub theme_error
 {
-&ui_print_header(undef, $text{'error'}, "");
+&header($text{'error'}, "");
+print "<p>\n";
 print "<h3>",($main::whatfailed ? "$main::whatfailed : " : ""),@_,"</h3>\n";
-&ui_print_footer();
+&footer();
 }
 
 sub theme_ui_print_unbuffered_header
